import asyncio
from logging import StreamHandler, Filter, Formatter, LogRecord
from asyncio import StreamWriter
from io import TextIOBase
from typing import Union, Type

from aiologger.protocols import AiologgerProtocol


class AsyncStreamHandler(StreamHandler):
    def __init__(self,
                 stream: StreamWriter,
                 level: Union[int, str],
                 formatter: Formatter,
                 filter: Filter = None):
        super().__init__(stream)
        self.setLevel(level)
        self.setFormatter(formatter)

        if filter:
            self.addFilter(filter)

    @classmethod
    async def init_from_pipe(cls, *,
                             pipe: TextIOBase,
                             level: Union[int, str],
                             formatter: Formatter,
                             filter: Filter = None,
                             protocol_factory: Type[asyncio.Protocol] = None,
                             loop=None) -> 'AsyncStreamHandler':
        if loop is None:
            loop = asyncio.get_event_loop()

        if protocol_factory is None:
            protocol_factory = AiologgerProtocol

        transport, protocol = await loop.connect_write_pipe(protocol_factory,
                                                            pipe)
        stream_writer = StreamWriter(transport=transport,
                                     protocol=protocol,
                                     reader=None,
                                     loop=loop)

        return AsyncStreamHandler(level=level,
                                  stream=stream_writer,
                                  formatter=formatter,
                                  filter=filter)

    async def handleError(self, record: LogRecord):
        """
        Handle errors which occur during an emit() call.

        This method should be called from handlers when an exception is
        encountered during an emit() call. If raiseExceptions is false,
        exceptions get silently ignored. This is what is mostly wanted
        for a logging system - most users will not care about errors in
        the logging system, they are more interested in application errors.
        You could, however, replace this with a custom handler if you wish.
        The record which was being processed is passed in to this method.
        """
        pass  # pragma: no cover

    async def handle(self, record: LogRecord) -> bool:
        """
        Conditionally emit the specified logging record.
        Emission depends on filters which may have been added to the handler.
        """
        rv = self.filter(record)
        if rv:
            await self.emit(record)
        return rv

    async def flush(self):
        await self.stream.drain()

    async def emit(self, record: LogRecord):
        """
        Actually log the specified logging record to the stream.
        """
        try:
            msg = self.format(record) + self.terminator

            await self.stream.write(msg.encode())
            await self.stream.drain()
        except Exception:
            await self.handleError(record)

    def close(self):
        """
        Tidy up any resources used by the handler.

        This version removes the handler from an internal map of handlers,
        _handlers, which is used for handler lookup by name. Subclasses
        should ensure that this gets called from overridden close()
        methods.
        """
        self.stream.close()



import json
import logging
import traceback
from datetime import datetime
from inspect import istraceback
from typing import Callable, Iterable
from datetime import timezone

LOGGED_AT_FIELDNAME = 'logged_at'
LINE_NUMBER_FIELDNAME = 'line_number'
FUNCTION_NAME_FIELDNAME = 'function'
LOG_LEVEL_FIELDNAME = 'level'
MSG_FIELDNAME = 'msg'
FILE_PATH_FIELDNAME = 'file_path'


class JsonFormatter(logging.Formatter):
    def __init__(self,
                 serializer: Callable[..., str] = json.dumps,
                 default_msg_fieldname: str = None):
        super(JsonFormatter, self).__init__()
        self.serializer = serializer

        self.default_msg_fieldname = default_msg_fieldname or MSG_FIELDNAME

    def _default_handler(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat()
        elif istraceback(obj):
            tb = ''.join(traceback.format_tb(obj))
            return tb.strip().split('\n')
        elif isinstance(obj, Exception):
            return "Exception: %s" % repr(obj)
        elif type(obj) is type:
            return str(obj)
        elif callable(obj):
            return obj()
        return str(obj)

    def format(self, record: logging.LogRecord) -> str:
        """
        Formats a record and serializes it as a JSON str. If record message isnt
        already a dict, initializes a new dict and uses `default_msg_fieldname`
        as a key as the record msg as the value.
        """
        msg = record.msg
        if not isinstance(record.msg, dict):
            msg = {self.default_msg_fieldname: msg}

        if record.exc_info:
            msg['exc_info'] = record.exc_info
        if record.exc_text:
            msg['exc_text'] = record.exc_text

        return self.serializer(msg, default=self._default_handler)


class ExtendedJsonFormatter(JsonFormatter):
    level_to_name_mapping = logging._levelToName
    default_fields = frozenset([
        LOG_LEVEL_FIELDNAME,
        LOGGED_AT_FIELDNAME,
        LINE_NUMBER_FIELDNAME,
        FUNCTION_NAME_FIELDNAME,
        FILE_PATH_FIELDNAME
    ])

    def __init__(self,
                 serializer: Callable[..., str] = json.dumps,
                 default_msg_fieldname: str = None,
                 exclude_fields: Iterable[str] = None,
                 tz: timezone = None):

        super(ExtendedJsonFormatter, self).__init__(
            serializer=serializer,
            default_msg_fieldname=default_msg_fieldname
        )
        self.tz = tz
        if exclude_fields is None:
            self.log_fields = self.default_fields
        else:
            self.log_fields = self.default_fields - set(exclude_fields)

    def formatter_fields_for_record(self, record):
        """
        :type record: aiologger.loggers.json.LogRecord
        """
        datetime_serialized = datetime.now(timezone.utc).astimezone(self.tz).isoformat()

        default_fields = (
            (LOGGED_AT_FIELDNAME, datetime_serialized),
            (LINE_NUMBER_FIELDNAME, record.lineno),
            (FUNCTION_NAME_FIELDNAME, record.funcName),
            (LOG_LEVEL_FIELDNAME, self.level_to_name_mapping[record.levelno]),
            (FILE_PATH_FIELDNAME, record.pathname)
        )

        for field, value in default_fields:
            if field in self.log_fields:
                yield field, value

    def format(self, record) -> str:
        """
        :type record: aiologger.loggers.json.LogRecord
        """
        msg = dict(self.formatter_fields_for_record(record))
        if record.flatten and isinstance(record.msg, dict):
            msg.update(record.msg)
        else:
            msg[MSG_FIELDNAME] = record.msg

        if record.extra:
            msg.update(record.extra)
        if record.exc_info:
            msg['exc_info'] = record.exc_info
        if record.exc_text:
            msg['exc_text'] = record.exc_text

        return self.serializer(msg,
                               default=self._default_handler,
                               **record.serializer_kwargs)



import asyncio
import logging
import sys
from asyncio import Lock, AbstractEventLoop, Task
from typing import Iterable, Optional, Callable, Awaitable, Tuple

from aiologger.filters import StdoutFilter
from aiologger.handlers import AsyncStreamHandler
from aiologger.protocols import AiologgerProtocol


_Caller = Tuple[str, int, str, Optional[str]]
_HandlerFactory = Callable[[], Awaitable[Iterable[logging.Handler]]]


class Logger(logging.Logger):
    def __init__(self, *,
                 name='aiologger',
                 level=logging.NOTSET,
                 loop=None,
                 formatter: Optional[logging.Formatter] = logging.Formatter,
                 handler_factory: Optional[_HandlerFactory] = None):
        super(Logger, self).__init__(name, level)
        self.loop: AbstractEventLoop = loop or asyncio.get_event_loop()
        self._handler_factory = handler_factory or (lambda: Logger._create_default_handlers(formatter, loop))
        self.initialized = False
        self._initializing = Lock()
        self._initialized = False
        self._was_shutdown = False

        async def _dummy(*args, **kwargs): return
        self.__dummy_task = self.loop.create_task(_dummy())

    @classmethod
    def with_default_handlers(cls, *, name='aiologger',
                              level=logging.NOTSET,
                              formatter: Optional[logging.Formatter] = None,
                              loop=None,
                              **kwargs):
        self = cls(name=name, level=level, loop=loop, formatter=formatter, **kwargs)

        return self

    @classmethod
    async def _create_default_handlers(cls,
                                       formatter: logging.Formatter = None,
                                       loop=None) -> Iterable[logging.Handler]:

        stdout_handler = await AsyncStreamHandler.init_from_pipe(
            pipe=sys.stdout,
            level=logging.DEBUG,
            protocol_factory=AiologgerProtocol,
            formatter=formatter,
            filter=StdoutFilter(),
            loop=loop)

        stderr_handler = await AsyncStreamHandler.init_from_pipe(
            pipe=sys.stderr,
            level=logging.WARNING,
            protocol_factory=AiologgerProtocol,
            formatter=formatter,
            loop=loop)
        return [stdout_handler, stderr_handler]

    async def _initialize(self):
        if not self._initialized:
            async with self._initializing:
                if self._initialized:
                    return
                for handler in await self._handler_factory():
                    self.addHandler(handler)
                self._initialized = True

    async def callHandlers(self, record):
        """
        Pass a record to all relevant handlers.

        Loop through all handlers for this logger and its parents in the
        logger hierarchy. If no handler was found, raises an error. Stop
        searching up the hierarchy whenever a logger with the "propagate"
        attribute set to zero is found - that will be the last logger
        whose handlers are called.
        """
        c = self
        found = 0
        while c:
            for handler in c.handlers:
                found = found + 1
                if record.levelno >= handler.level:
                    await handler.handle(record)
            if not c.propagate:
                c = None  # break out
            else:
                c = c.parent
        if found == 0:
            raise Exception("No handlers could be found for logger")

    async def handle(self, record):
        """
        Call the handlers for the specified record.

        This method is used for unpickled records received from a socket, as
        well as those created locally. Logger-level filtering is applied.
        """
        await self._initialize()

        if (not self.disabled) and self.filter(record):
            await self.callHandlers(record)

    async def _log(self,
                   level,
                   msg,
                   args,
                   exc_info=None,
                   extra=None,
                   stack_info=False,
                   caller: _Caller = None):

        sinfo = None
        if logging._srcfile and caller is None:
            # IronPython doesn't track Python frames, so findCaller raises an
            # exception on some versions of IronPython. We trap it here so that
            # IronPython can use logging.
            try:
                fn, lno, func, sinfo = self.findCaller(stack_info)
            except ValueError:  # pragma: no cover
                fn, lno, func = "(unknown file)", 0, "(unknown function)"
        elif caller:
            fn, lno, func, sinfo = caller
        else:  # pragma: no cover
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info and isinstance(exc_info, BaseException):
            exc_info = (type(exc_info), exc_info, exc_info.__traceback__)

        record = logging.LogRecord(
            name=self.name,
            level=level,
            pathname=fn,
            lineno=lno,
            msg=msg,
            args=args,
            exc_info=exc_info,
            func=func,
            sinfo=sinfo,
            extra=extra
        )
        await self.handle(record)

    def _make_log_task(self, level, msg, *args, **kwargs) -> Task:
        """
        Creates an asyncio.Task for a msg if logging is enabled for level.
        Returns a dummy task otherwise.
        """
        if not self.isEnabledFor(level):
            return self.__dummy_task

        if kwargs.get('exc_info', False):
            if not isinstance(kwargs['exc_info'], BaseException):
                kwargs['exc_info'] = sys.exc_info()

        coro = self._log(level, msg, *args,
                         caller=self.findCaller(False),
                         **kwargs)
        return self.loop.create_task(coro)

    def debug(self, msg, *args, **kwargs) -> Task:
        """
        Log msg with severity 'DEBUG'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        await logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        return self._make_log_task(logging.DEBUG, msg, args, **kwargs)

    def info(self, msg, *args, **kwargs) -> Task:
        """
        Log msg with severity 'INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        await logger.info("Houston, we have an interesting problem", exc_info=1)
        """
        return self._make_log_task(logging.INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs) -> Task:
        """
        Log msg with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        await logger.warning("Houston, we have a bit of a problem", exc_info=1)
        """
        return self._make_log_task(logging.WARNING, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs) -> Task:
        """
        Log msg with severity 'ERROR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        await logger.error("Houston, we have a major problem", exc_info=1)
        """
        return self._make_log_task(logging.ERROR, msg, args, **kwargs)

    def critical(self, msg, *args, **kwargs) -> Task:
        """
        Log msg with severity 'CRITICAL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        await logger.critical("Houston, we have a major disaster", exc_info=1)
        """
        return self._make_log_task(logging.CRITICAL, msg, args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs) -> Task:
        """
        Convenience method for logging an ERROR with exception information.
        """
        return self.error(msg, *args, exc_info=exc_info, **kwargs)

    async def shutdown(self):
        """
        Perform any cleanup actions in the logging system (e.g. flushing
        buffers).

        Should be called at application exit.
        """
        if self._was_shutdown:
            return
        self._was_shutdown = True

        for handler in reversed(self.handlers):
            if not handler:
                continue
            try:
                if self._initialized:
                    await handler.flush()
                    handler.close()

            except Exception:
                """
                Ignore errors which might be caused
                because handlers have been closed but
                references to them are still around at
                application exit. Basically ignore everything, 
                as we're shutting down
                """
                pass



import json
import logging
import sys
from datetime import timezone
from asyncio import AbstractEventLoop
from typing import Dict, Iterable, Callable, Tuple, Any, Optional, Awaitable

from aiologger import Logger
from aiologger.formatters.json import ExtendedJsonFormatter
from aiologger.logger import _Caller


class LogRecord(logging.LogRecord):
    def __init__(self, name, level, pathname, lineno,
                 msg, args, exc_info, func=None, sinfo=None, **kwargs):
        super().__init__(name, level, pathname, lineno, msg,
                         args, exc_info, func, sinfo)
        self.extra = kwargs['extra']
        self.flatten = kwargs['flatten']
        self.serializer_kwargs = kwargs['serializer_kwargs']


class JsonLogger(Logger):
    def __init__(self,
                 name: str = 'aiologger-json',
                 level: int = logging.DEBUG,
                 serializer: Callable[..., str] = json.dumps,
                 flatten: bool = False,
                 serializer_kwargs: Dict = None,
                 extra: Dict = None,
                 exclude_fields: Iterable[str] = None,
                 loop: AbstractEventLoop = None,
                 tz: timezone = None,
                 formatter: Optional[logging.Formatter] = None,
                 handler_factory: Optional[Callable[[], Awaitable[Iterable[logging.Handler]]]] = None):
        formatter = formatter or ExtendedJsonFormatter(serializer=serializer,
                                                       exclude_fields=exclude_fields,
                                                       tz=tz)
        super().__init__(name=name, level=level, loop=loop, formatter=formatter, handler_factory=handler_factory)

        self.flatten = flatten

        if serializer_kwargs is None:
            serializer_kwargs = {}
        self.serializer_kwargs = serializer_kwargs

        if extra is None:
            extra = {}
        self.extra = extra

    @classmethod
    def with_default_handlers(cls, *,
                              name: str = 'aiologger-json',
                              level: int = logging.NOTSET,
                              serializer: Callable[..., str] = json.dumps,
                              flatten: bool = False,
                              serializer_kwargs: Dict = None,
                              extra: Dict = None,
                              exclude_fields: Iterable[str] = None,
                              loop: AbstractEventLoop = None,
                              tz: timezone = None):
        return super(JsonLogger, cls).with_default_handlers(
            name='aiologger-json',
            level=level,
            loop=loop,
            serializer=serializer,
            flatten=flatten,
            serializer_kwargs=serializer_kwargs,
            extra=extra,
            exclude_fields=exclude_fields,
            tz=tz
        )

    async def _log(self,
                   level: int,
                   msg: Any,
                   args: Tuple,
                   exc_info=None,
                   extra: Dict = None,
                   stack_info=False,
                   flatten: bool = False,
                   serializer_kwargs: Dict = None,
                   caller: _Caller=None):
        """
        Low-level logging routine which creates a LogRecord and then calls
        all the handlers of this logger to handle the record.

        Overwritten to properly handle log methods kwargs
        """
        sinfo = None
        if caller:
            fn, lno, func, sinfo = caller
        else:  # pragma: no cover
            fn, lno, func = "(unknown file)", 0, "(unknown function)"
        if exc_info and isinstance(exc_info, BaseException):
            exc_info = (type(exc_info), exc_info, exc_info.__traceback__)

        joined_extra = {}
        joined_extra.update(self.extra)

        if extra:
            joined_extra.update(extra)

        record = LogRecord(
            name=self.name,
            level=level,
            pathname=fn,
            lineno=lno,
            msg=msg,
            args=args,
            exc_info=exc_info,
            func=func,
            sinfo=sinfo,
            extra=joined_extra,
            flatten=flatten or self.flatten,
            serializer_kwargs=serializer_kwargs or self.serializer_kwargs
        )
        await self.handle(record)



from setuptools import setup, find_packages


VERSION = '0.2.3-rc1'
setup(
    name='aiologger',
    version=VERSION,
    packages=find_packages(exclude=['*test*']),
    url='https://github.com/diogommartins/aiologger',
    author='Diogo Magalhães Martins',
    author_email='magalhaesmartins@icloud.com',
    keywords='logging json log output',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: System :: Logging'
    ]
)
